# Amazon Web Services, Inc. or its affiliates. All Rights Reserved. This AWS Content
# is provided subject to the terms of the AWS Customer Agreement available at 
# https://aws.amazon.com/agreement/ or other written agreement between Customer
# and Amazon Web Services, Inc.

######
# Checks for compliance on activities that could generate violations
# Checks for the use of TLS Termination on the ELBs based off of the following API calls.
# API Triggers: CreateLoadBalancer, CreateLoadBalancerListeners,
#       DeleteLoadBalancerListeners, CreateLoadBalancerPolicy, 
#       SetLoadBalancerPoliciesOfListener, CreateListener, 
#       ModifyListener, DeleteListener
# Services: Cloudwatch Events (trigger), Lambda, Classic Loadbalancer, 
#           Network LoadBalancer, Application LoadBalancer
# Python 3.6 - Lambda 
######
# Begin imports and client declarations
import boto3
import json
import logging
import os
from botocore.exceptions import ClientError

client_elb = boto3.client('elb')
client_elbv2 = boto3.client('elbv2')
try:
    OUTBOUND_TOPIC_ARN = os.environ["outbound_topic_arn"]
except AttributeError as e:
    print(e)
    exit(1)

# Main Lambda handler
def lambda_handler(event, context):
    setup_logging()

    policy_list = []
    log.info(json.dumps(event))
    eval_results = []
    if event["detail"]["eventName"] in ["CreateLoadBalancerListeners", "DeleteLoadBalancerListeners",\
         "CreateLoadBalancer", "DeleteListener"]:
        # For Application ELB
        if "type" in event["detail"]["requestParameters"]:
            if event["detail"]["requestParameters"]["type"] == "application":
                eval_results.append(get_application_protocol_violations(event))
        if event["detail"]["eventName"] == "DeleteListener":
            eval_results.append(get_application_protocol_violations(event))
        # Classic ELB
        if "loadBalancerName" in event["detail"]["requestParameters"]:
            protocols = check_classic_listener(event)
            eval_results.append(get_classic_protocol_violations(event, protocols))

    if event["detail"]["eventName"] == "CreateLoadBalancerPolicy":
        protocols = get_classic_loadbalancer_protocols(event)
        eval_results.append(get_classic_protocol_violations(event, protocols))

    if event["detail"]["eventName"] == "SetLoadBalancerPoliciesOfListener":
        if "policyNames" in event["detail"]["requestParameters"]:
            policy_list.append(event["detail"]["requestParameters"]["policyNames"])
            log.info(policy_list)
            LoadBalancerName = event["detail"]["requestParameters"]["loadBalancerName"]
        protocols = describe_classic_loadbalancer_policies(LoadBalancerName, policy_list)
        eval_results.append(get_classic_protocol_violations(event, protocols))

    # Application ELB Section:
    if event["detail"]["eventName"] in ["CreateListener", "ModifyListener" ]:
        eval_results.append(get_application_protocol_violations(event))

    # Check if a violation was had
    if False in eval_results:
        subject = "Violation - ELB Does not Terminate TLS!"
        message = create_non_compliance_message(event, context)
        send_violation(message, subject)

# Describe Loadbalancers for the Classic Loadbalancers
def describe_classic_loadbalancers(LoadBalancerName):
    protocols = []
    protocol_list = []
    policy_list = []
    try:
        response = client_elb.describe_load_balancers(LoadBalancerNames=[LoadBalancerName,],)
        if "ListenerDescriptions" in response["LoadBalancerDescriptions"][0]:
            for listeners in response["LoadBalancerDescriptions"][0]["ListenerDescriptions"]:
                protocol_list.append(listeners["Listener"]["Protocol"])
                if listeners["Listener"]["Protocol"] != "HTTP":
                    policy_list.append(listeners["PolicyNames"])
            log.info(protocol_list)
            if ("HTTPS" not in protocol_list) and ("SSL" not in protocol_list):
                log.warning('Neither HTTPS or SSL is configured on this load balancer.')
                log.warning(protocol_list)
                protocols.append({"MissingPolicy": 'true'})
                return policy_list, protocols
                
    except ClientError as err:
        log.error(err)
    return policy_list, protocols

# Get SSL protocols that are enabled on the ELB
def check_classic_listener(event):
    if "loadBalancerName" in event["detail"]["requestParameters"]:
        LoadBalancerName = event["detail"]["requestParameters"]["loadBalancerName"]
        policy_list, protocols = describe_classic_loadbalancers(LoadBalancerName)
        if policy_list:
            protocols = describe_classic_loadbalancer_policies(LoadBalancerName, policy_list)
            log.info(protocols)
        return protocols

# Describe Classic Load Balancer policies
def describe_classic_loadbalancer_policies(LoadBalancerName, policy_list):
    protocols = []
    for policy in policy_list:
        try:
            response = client_elb.describe_load_balancer_policies(LoadBalancerName=LoadBalancerName,\
            PolicyNames=[policy[0],])
            if "PolicyAttributeDescriptions" in response["PolicyDescriptions"][0]:
                for policy_attribute in response["PolicyDescriptions"][0]\
                    ["PolicyAttributeDescriptions"]:
                    if policy_attribute["AttributeName"].startswith("Protocol-"):
                        protocols.append({policy_attribute["AttributeName"]: \
                        policy_attribute["AttributeValue"]})
        except ClientError as err:
            log.error(err)
    return protocols

# Gets Application Protocol Violations
def get_application_protocol_violations(event):
    #DeleteListener requires special handling to determine the load balancer ARN.
    if event["detail"]["eventName"] == "DeleteListener":
        temp_alb = event["detail"]["requestParameters"]["listenerArn"].split("/")
        temp_alb2 = event["detail"]["requestParameters"]["listenerArn"].split(":")
        temp_alb = ':loadbalancer/app/' + temp_alb[2] + '/' + temp_alb[3]
        temp_alb2 = temp_alb2[0] + ':' + temp_alb2[1]+ ':' + temp_alb2[2]+ ':' + \
        temp_alb2[3]+ ':' + temp_alb2[4]
        loadBalancerArn = temp_alb2 + temp_alb
        log.info(loadBalancerArn)
        ssl_policy = describe_application_loadbalancer_listeners(loadBalancerArn)

    if event["detail"]["eventName"] in ["ModifyListener", "CreateListener"]:
        if "errorCode" in event["detail"]:
            ssl_policy = False
            log.error('Server side error code was returned by AWS: ' + event["detail"]\
            ["errorMessage"])
            return ssl_policy
        loadBalancerArn = event["detail"]["responseElements"]["listeners"][0]["loadBalancerArn"]
        ssl_policy = describe_application_loadbalancer_listeners(loadBalancerArn)

    if event["detail"]["eventName"] == "CreateLoadBalancer": 
        if "loadBalancers" in event["detail"]["responseElements"]:
            loadBalancerArn = event["detail"]["responseElements"]["loadBalancers"][0]\
            ["loadBalancerArn"]
            ssl_policy = describe_application_loadbalancer_listeners(loadBalancerArn)
            
    if ssl_policy == False:
        log.info(ssl_policy)
        log.warning('Either no HTTPS listener was detected, or something unexpected happened \
        with ssl_policy in get_application_protocol_violations.')
        return False     

    for policy in ssl_policy:
        if policy != "ELBSecurityPolicy-TLS-1-2-2017-01":
            log.warning("Violation - You are using policy " + policy + \
            " Only ELBSecurityPolicy-TLS-1-2-2017-01 is allowed for HTTPS")
            return False
    log.info('HTTPS Listener is utilizing ELBSecurityPolicy-TLS-1-2-2017-01. \
    Nothing to do here.')
    return True

# Describe Listeners for the Application LoadBalancer
def describe_application_loadbalancer_listeners(loadBalancerArn):
    try:
        ssl_policy = []
        response = client_elbv2.describe_listeners(LoadBalancerArn=loadBalancerArn,)
        log.info(json.dumps(response))
        listeners = response["Listeners"]
        for listener in listeners:
            if "SslPolicy" in listener:
                ssl_policy.append(listener["SslPolicy"])
                log.info('ssl_policy in describe_application_loadbalancer_listeners is ' +\
                 str(ssl_policy))
        log.info(ssl_policy)
        if ssl_policy:
            return ssl_policy
        else:
            log.warning('No HTTPS/SSL Policy was found in \
            describe_application_loadbalancer_listeners')
            return False
    except ClientError as err:
        log.error(err)
        return False
   
# Decide whether classic ELB violates our rules.
def get_classic_protocol_violations(event, protocols):
    log.info(protocols)
    if protocols:
        for protocol in protocols:
            log.info(protocol)
            for key, value in protocol.items():
                if key != "Protocol-TLSv1.2" and value == 'true':
                    log.warning( 'This load balancer is missing a valid protocol or policy: '\
                     + key)
                    return False
    if not protocols:
        log.warning('Detected empty protocols list in get_classic_protocol_violations')
        return False   
    log.info('We did not detect any problems with the load balancers configuration.')
    return True
    
#Describe the SSL Protocols in use by the policy assigned to the SSL listener on ELB
def get_classic_loadbalancer_protocols(event):
    protocols = []
    policy_list = []
    policy = []
    if "policyAttributes" in event["detail"]["requestParameters"]:
        for policy_attribute in event["detail"]["requestParameters"]["policyAttributes"]:
            if policy_attribute["attributeName"].startswith( "Protocol-" ):
                protocols.append({policy_attribute["attributeName"]: \
                policy_attribute["attributeValue"]})
            if policy_attribute["attributeValue"].startswith("ELBSecurityPolicy-"):
                policy.append(event["detail"]["requestParameters"]["policyName"])
                log.info(policy)
                policy_list.append(policy)
                log.info(policy_list)
                LoadBalancerName = event["detail"]["requestParameters"]["loadBalancerName"]
                protocols = describe_classic_loadbalancer_policies(LoadBalancerName, policy_list)
                log.info(protocols)
    return protocols

def send_violation(message, subject):
    """
    Function will publish the AWS SNS notification detailing when a violation has occurred.
    """
    findsnsregion = OUTBOUND_TOPIC_ARN.split(":")
    snsregion = findsnsregion[3]
    sendclient = boto3.client('sns', region_name=snsregion)
    try:
        sendclient.publish(
            TopicArn=OUTBOUND_TOPIC_ARN,
            Message=message,
            Subject=subject
        )
    except ClientError as err:
        log.info(err)
        return False


def create_non_compliance_message(event, context):
    """
    Function will structure the outgoing AWS SNS notification format
    """
    message = "Violation - ELB does not terminate TLS!  \n\n"
    message += 'Full event which caused this error:\n\n'
    message += json.dumps(event, sort_keys=True,
                  indent=4, separators=(',', ': '))
    message += "\n\n"
    message += "This notification was generated by the Lambda function " + \
        context.invoked_function_arn
    return message

def setup_logging():
    """
    Logging Function.

    Creates a global log object and sets its level.
    """
    global log
    log = logging.getLogger()
    log_levels = {'INFO': 20, 'WARNING': 30, 'ERROR': 40}

    if 'logging_level' in os.environ:
        log_level = os.environ['logging_level'].upper()
        if log_level in log_levels:
            log.setLevel(log_levels[log_level])
        else:
            log.setLevel(log_levels['ERROR'])
            log.error("The logging_level environment variable is not set to INFO, WARNING, or \
                        ERROR.  The log level is set to ERROR")
    else:
        log.setLevel(log_levels['ERROR'])
    log.info('Logging setup complete - set to log level ' + str(log.getEffectiveLevel()))